# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: grpc.proto
# Protobuf Python Version: 5.29.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    0,
    '',
    'grpc.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\ngrpc.proto\x12\tremote_rf\"\xa2\x01\n\x11GenericRPCRequest\x12\x15\n\rfunction_name\x18\x01 \x01(\t\x12\x34\n\x04\x61rgs\x18\x02 \x03(\x0b\x32&.remote_rf.GenericRPCRequest.ArgsEntry\x1a@\n\tArgsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\"\n\x05value\x18\x02 \x01(\x0b\x32\x13.remote_rf.Argument:\x02\x38\x01\"\x96\x01\n\x12GenericRPCResponse\x12;\n\x07results\x18\x01 \x03(\x0b\x32*.remote_rf.GenericRPCResponse.ResultsEntry\x1a\x43\n\x0cResultsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\"\n\x05value\x18\x02 \x01(\x0b\x32\x13.remote_rf.Argument:\x02\x38\x01\"\x19\n\nArrayShape\x12\x0b\n\x03\x64im\x18\x01 \x03(\x05\"+\n\rComplexNumber\x12\x0c\n\x04real\x18\x01 \x01(\x02\x12\x0c\n\x04imag\x18\x02 \x01(\x02\"a\n\x11\x43omplexNumpyArray\x12$\n\x05shape\x18\x01 \x01(\x0b\x32\x15.remote_rf.ArrayShape\x12&\n\x04\x64\x61ta\x18\x02 \x03(\x0b\x32\x18.remote_rf.ComplexNumber\"D\n\x0eRealNumpyArray\x12$\n\x05shape\x18\x01 \x01(\x0b\x32\x15.remote_rf.ArrayShape\x12\x0c\n\x04\x64\x61ta\x18\x02 \x03(\x02\"\xd7\x01\n\x08\x41rgument\x12\x16\n\x0cstring_value\x18\x01 \x01(\tH\x00\x12\x15\n\x0bint64_value\x18\x02 \x01(\x03H\x00\x12\x15\n\x0b\x66loat_value\x18\x03 \x01(\x02H\x00\x12\x14\n\nbool_value\x18\x04 \x01(\x08H\x00\x12\x35\n\rcomplex_array\x18\x05 \x01(\x0b\x32\x1c.remote_rf.ComplexNumpyArrayH\x00\x12/\n\nreal_array\x18\x06 \x01(\x0b\x32\x19.remote_rf.RealNumpyArrayH\x00\x42\x07\n\x05value2Q\n\nGenericRPC\x12\x43\n\x04\x43\x61ll\x12\x1c.remote_rf.GenericRPCRequest\x1a\x1d.remote_rf.GenericRPCResponseB\x1e\n\x10\x63om.example.demoB\nDemoProtosb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'grpc_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\020com.example.demoB\nDemoProtos'
  _globals['_GENERICRPCREQUEST_ARGSENTRY']._loaded_options = None
  _globals['_GENERICRPCREQUEST_ARGSENTRY']._serialized_options = b'8\001'
  _globals['_GENERICRPCRESPONSE_RESULTSENTRY']._loaded_options = None
  _globals['_GENERICRPCRESPONSE_RESULTSENTRY']._serialized_options = b'8\001'
  _globals['_GENERICRPCREQUEST']._serialized_start=26
  _globals['_GENERICRPCREQUEST']._serialized_end=188
  _globals['_GENERICRPCREQUEST_ARGSENTRY']._serialized_start=124
  _globals['_GENERICRPCREQUEST_ARGSENTRY']._serialized_end=188
  _globals['_GENERICRPCRESPONSE']._serialized_start=191
  _globals['_GENERICRPCRESPONSE']._serialized_end=341
  _globals['_GENERICRPCRESPONSE_RESULTSENTRY']._serialized_start=274
  _globals['_GENERICRPCRESPONSE_RESULTSENTRY']._serialized_end=341
  _globals['_ARRAYSHAPE']._serialized_start=343
  _globals['_ARRAYSHAPE']._serialized_end=368
  _globals['_COMPLEXNUMBER']._serialized_start=370
  _globals['_COMPLEXNUMBER']._serialized_end=413
  _globals['_COMPLEXNUMPYARRAY']._serialized_start=415
  _globals['_COMPLEXNUMPYARRAY']._serialized_end=512
  _globals['_REALNUMPYARRAY']._serialized_start=514
  _globals['_REALNUMPYARRAY']._serialized_end=582
  _globals['_ARGUMENT']._serialized_start=585
  _globals['_ARGUMENT']._serialized_end=800
  _globals['_GENERICRPC']._serialized_start=802
  _globals['_GENERICRPC']._serialized_end=883
# @@protoc_insertion_point(module_scope)
